/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.beam.sdk.util.common;

import static org.apache.beam.sdk.util.common.Counter.AggregationKind.MEAN;

import org.apache.beam.sdk.coders.Coder;
import org.apache.beam.sdk.util.common.Counter.CounterMean;

import org.junit.Assert;

import java.io.ByteArrayOutputStream;

/**
 * Utilities for testing {@link Counter}s.
 */
public class CounterTestUtils {
  /**
   * A utility method that passes the given (unencoded) elements through
   * coder's registerByteSizeObserver() and encode() methods, and confirms
   * they are mutually consistent. This is useful for testing coder
   * implementations.
   */
  public static <T> void testByteCount(Coder<T> coder, Coder.Context context, T[] elements)
      throws Exception {
    Counter<Long> meanByteCount = Counter.longs("meanByteCount", MEAN);
    ElementByteSizeObserver observer = new ElementByteSizeObserver(meanByteCount);

    ByteArrayOutputStream os = new ByteArrayOutputStream();
    for (T elem : elements) {
      coder.registerByteSizeObserver(elem, observer, context);
      coder.encode(elem, os, context);
      observer.advance();
    }
    long expectedLength = os.toByteArray().length;

    CounterMean<Long> mean = meanByteCount.getMean();

    Assert.assertEquals(expectedLength, mean.getAggregate().longValue());
    Assert.assertEquals(elements.length, mean.getCount());
  }
}
